/* Copyright (C) 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package backend.srb;

import memory.MemoryImage;
import osio.TextPrinter;
import backend.Instruction;
import backend.LittleEndian;

import compbase.Context;
import compbase.Marks;
import compbase.Mthd;
import compbase.StdTypes;

/**
 * SRB: backend for optimized code generation 
 * 
 * @author S. Frenz
 * @version 090219 adopted changed Architecture (genCopyInstContext<->getClssInstReg)
 *  version 090214 prepared for native register allocation
 *  version 090208 removed genClearMem
 *  version 090207 added copyright notice and removed genLoadNullAddr/genPushNull
 *  version 090118 added register count functionallity and printCode
 *  version 081226 initial version
 */

public class SRB extends LittleEndian {
  /**
   * register definition
   * 1 -> class context
   * 2 -> instance context (if existing)
   * 3 -> base pointer
   * 4 -> stack pointer
   * 5 -> primary (result) register
   * 6... general purpose registers
   */
  public final static int R_CLSS = 1;
  public final static int R_INST = 2;
  public final static int R_BASE = 3;
  public final static int R_PRIR = 4;
  public final static int R_GPRS = 5;
  
  /**
   * instruction type definition
   */
  public final static int I_JUMP            = -9;
  public final static int I_MTHDFRAME       =  1;
  public final static int I_FLOWHINT        =  2;
  public final static int I_LOADCONSTVAL    =  3;
  public final static int I_LOADVAR         =  4;
  public final static int I_DUPCONV         =  5;
  public final static int I_SINGLESTACKOP   =  6;
  public final static int I_ASSIGN          =  7;
  public final static int I_REGOP           =  8;
  public final static int I_MEMOP           =  9;
  public final static int I_CALL            = 10;
  public final static int I_COMPARE         = 11;
  public final static int I_SPECIALMEMIO    = 12;
  public final static int I_DEREFADDR       = 13;
  public final static int I_CHECKBOUNDS     = 14;
  public final static int I_NATIVEBOUNDEXC  = 15;
  public final static int I_THROWFRAME      = 16;
  public final static int I_INLINEVAROFFSET = 17;
  
  /**
   * additional register type as supporting register
   */
  public final static int T_SPRT = -101;
  
  /**
   * phases of native instruction handling
   */
  public final static int P_DECORATEREGISTERS   = 1;
  public final static int P_RESOLVEREGISTERS    = 2;
  public final static int P_MAPSUPPORTREGISTERS = 3;
  public final static int P_GENERATECODE        = 4;
  
  //internal variables
  private SRBI_Container currentContainer, startContainer;
  private Mthd currentMthd;
  protected Mthd outestMthd;
  protected SRBTarget target;
  
  protected SRBI_Single freeInstruction;
  protected SRBI_Container freeContainer;
  protected SRBRegister firstGPReg, lastReg, freeReg; //first / last allocated and first empty register
  protected SRBConstraint firstConstraint, freeConstraint;
  private int nextID, nextReg, nextAlloc;
  private int statNewIns, statNewCont, statNewReg;
  
  public SRB(SRBTarget it) {
    target=it;
  }
  
  //methods for SRB specials
  
  public void addConstraint(int type, SRBRegister reg, boolean recurseSupporting) {
    addConstraint(type, reg, null, 0, 0l, recurseSupporting);
  }
  
  public void addConstraint(int type, SRBRegister reg1, SRBRegister reg2, boolean recurseSupporting) {
    addConstraint(type, reg1, reg2, 0, 0l, recurseSupporting);
  }
  
  public void addConstraint(int type, SRBRegister reg, int constInt, boolean recurseSupporting) {
    addConstraint(type, reg, null, constInt, 0l, recurseSupporting);
  }
  
  public void addConstraint(int type, SRBRegister reg, long constLong, boolean recurseSupporting) {
    addConstraint(type, reg, null, 0, constLong, recurseSupporting);
  }
  
  public void addConstraint(int type, SRBRegister reg1, SRBRegister reg2, int constInt, long constLong,
      boolean recurseSupporting) {
    SRBConstraint c;
    if ((c=freeConstraint)==null) c=new SRBConstraint();
    else freeConstraint=freeConstraint.nextConstraint;
    c.nextConstraint=firstConstraint;
    c.type=type;
    c.reg1=reg1;
    c.reg2=reg2;
    c.constInt=constInt;
    c.constLong=constLong;
    firstConstraint=c;
    if (recurseSupporting) {
      if (reg1!=null && reg1.containerNext!=null && reg1.containerNext.type==T_SPRT)
        addConstraint(type, reg1.containerNext, reg2, constInt, constLong, recurseSupporting);
      if (reg2!=null && reg2.containerNext!=null && reg2.containerNext.type==T_SPRT)
        addConstraint(type, reg1, reg2.containerNext, constInt, constLong, recurseSupporting);
    }
  }
  
  public SRBConstraint getNextConstraint(int type, SRBRegister searchReg, SRBConstraint startSearch) {
    if (startSearch==null) startSearch=firstConstraint;
    else startSearch=startSearch.nextConstraint;
    while (startSearch!=null) {
      if ((startSearch.reg1==searchReg || startSearch.reg2==searchReg)
          && (type==SRBConstraint.C_ALL || startSearch.type==type)) return startSearch;
      startSearch=startSearch.nextConstraint;
    }
    return null;
  }
  
  //methods to fulfill Architecture
  
  public void init(MemoryImage imem, int ilev, Context ictx) {
    super.init(imem, ilev, ictx);
    target.init(imem, ilev, ictx);
    target.backInit(this);
  }
  
  public void printStatistics() {
    ctx.out.print("objects created for ins / cont / reg: ");
    ctx.out.print(statNewIns);
    ctx.out.print(" / ");
    ctx.out.print(statNewCont);
    ctx.out.print(" / ");
    ctx.out.println(statNewReg);
  }
  
  public void putRef(int pos, int ptr) {
    target.putRef(pos, ptr);
  }
  
  public void putCodeRef(int pos, int ptr) {
    target.putCodeRef(pos, ptr);
  }
  
  public void appendInstruction(Instruction i) { //override normal instruction structure
    currentContainer.addBlock(i);
  }
  
  public void finalizeInstructions(Instruction first) {
    if (ctx.printCode || (outestMthd.marker&Marks.K_PRCD)!=0) printCode(ctx.out, "SRB-start");
    startContainer.generateJumpDestsForTarget();
    //TODO optimize SRB-blocks
    startContainer.convertToNative(P_DECORATEREGISTERS);
    //TODO optimize decorated instructions
    //TODO optimize usage of native registers
    startContainer.convertToNative(P_RESOLVEREGISTERS);
    startContainer.convertToNative(P_MAPSUPPORTREGISTERS);
    if (ctx.printCode || (outestMthd.marker&Marks.K_PRCD)!=0) printCode(ctx.out, "SRB-optimized");
    startContainer.convertToNative(P_GENERATECODE);
    target.finalizeMethodCoding();
    err|=target.hasError();
    //clean up
    startContainer.cleanUp();
  }
  
  public int getMethodSize() {
    return target.getMethodSize();
  }
  
  public void copyMethodCode(int destination) {
    target.copyMethodCode(destination);
  }
  
  public Instruction createNewInstruction() {
    statNewIns++;
    return new SRBI_Single(this, maxInstrCodeSize);
  }
  
  public Instruction getUnlinkedInstruction() {
    Instruction i;
    if ((i=freeInstruction)==null) i=createNewInstruction();
    else freeInstruction=(SRBI_Single)freeInstruction.next;
    i.type=I_NONE;
    i.reg0=i.reg1=i.reg2=i.size=0;
    i.iPar1=i.iPar2=i.iPar3=0;
    i.lPar=0l;
    i.jDest=null;
    i.isDest=false;
    i.refMthd=null;
    i.prev=i.next=null;
    return i;
  }
  
  public SRBI_Container getUnlinkedContainer() {
    SRBI_Container c;
    if ((c=freeContainer)==null) {
      statNewCont++;
      c=new SRBI_Container(this);
    }
    else freeContainer=(SRBI_Container)freeContainer.next;
    c.outer=null;
    c.next=null;
    return c;
  }
  
  public Mthd codeProlog(Mthd mthd) {
    Mthd lastMthd;
    SRBI_Container lastContainer;
    SRBConstraint lastConstraint;
    
    lastContainer=currentContainer;
    currentContainer=getUnlinkedContainer();
    if (lastContainer==null) {
      nextID=1;
      nextReg=R_GPRS;
      startContainer=currentContainer;
      outestMthd=mthd;
      //allocate special registers and fixate native registers
      enableSpecialRegister(R_CLSS);
      enableSpecialRegister(R_INST);
      enableSpecialRegister(R_BASE);
      //reset register and constraint lists
      firstGPReg=null; //reset first general purpose register
      if ((lastConstraint=firstConstraint)!=null) {
        while (lastConstraint.nextConstraint!=null) lastConstraint=lastConstraint.nextConstraint;
        lastConstraint.nextConstraint=freeConstraint;
        freeConstraint=firstConstraint;
        firstConstraint=null;
      }
    }
    else lastContainer.addBlock(currentContainer);
    insMthdFrame(1, mthd);
    lastMthd=currentMthd;
    currentMthd=mthd;
    return lastMthd;
  }
  
  private void enableSpecialRegister(int regNr) {
    enableRegister(regNr, StdTypes.T_PTR);
    target.initSpecialRegister(lastReg);
  }
  
  public void codeEpilog(Mthd outline) {
    insMthdFrame(0, outline);
    if (currentContainer.outer==null) {
      deallocRestoreReg(R_BASE, 0, 0);
      deallocRestoreReg(R_INST, 0, 0);
      deallocRestoreReg(R_CLSS, 0, 0);
      currentContainer=null;
    }
    else currentContainer=currentContainer.outer;
    currentMthd=outline;
  }
  
  public int ensureFreeRegs(int ignoreReg1, int ignoreReg2, int keepReg1, int keepReg2) {
    int cnt=0;
    SRBRegister reg=lastReg, lSaved=null;
    
    while (reg!=null) {
      if (reg.saveStopper>0 || reg.nr<R_GPRS) break;
      if (reg.nr!=ignoreReg1 && reg.nr!=ignoreReg2) {
        insSingleStackOp(1, reg.nr, 0, 0l, reg.type);
        reg.saved++;
      }
      else {
        if (reg.saved!=0) {
          ctx.out.println("SRB: can not ignore already saved register");
          err=true;
        }
        //ensure without push
      }
      cnt++;
      lSaved=reg;
      reg=reg.allocPrev;
    }
    if (lSaved!=null) lSaved.saveStopper++;
    return cnt;
  }
  
  public int prepareFreeReg(int avoidReg1, int avoidReg2, int reUseReg, int type) {
    int cnt, nr;
    
    //never reuse a register, so there is never a conflict: avoidReg1/2 not relevant
    nextAlloc=nextReg++;
    //enter register(s) in used register chain
    enableRegister(nr=nextAlloc, type);
    cnt=target.getRegisterCount(type);
    while (--cnt>0) {
      enableRegister(++nr, T_SPRT);
      ++nextReg; //register used as support, next register is this+1
    }
    //return: nothing to restore
    return 0; //nothing has to be freed
  }
  
  private void enableRegister(int nr, int type) {
    SRBRegister reg;
    
    if (freeReg!=null) {
      reg=freeReg;
      freeReg=freeReg.allocNext;
      reg.containerNext=reg.allocPrev=reg.allocNext=null;
    }
    else {
      reg=new SRBRegister();
      reg.nativeInfo=target.createNativeRegisterInfo();
      statNewReg++;
    }
    currentContainer.openReg(reg);
    if ((reg.nr=nr)==R_GPRS) firstGPReg=reg;
    reg.type=type;
    reg.saveStopper=reg.saved=0;
    reg.nativeInfo.reset();
    if (lastReg==null) lastReg=reg;
    else {
      lastReg.allocNext=reg;
      reg.allocPrev=lastReg;
      lastReg=reg;
    }
  }
  
  public int allocReg() {
    return nextAlloc;
  }
  
  public void deallocRestoreReg(int deallocRegs, int keepRegs, int restoreRegs) {
    SRBRegister cmp;
    
    //restore register
    if (restoreRegs!=0) {
      cmp=lastReg;
      //search saveStopper
      while (cmp!=null && cmp.saveStopper==0) cmp=cmp.allocPrev;
      if (cmp==null) {
        ctx.out.println("SRB: invalid saveStopper (null)");
        err=true;
      }
      else { //restore registers
        cmp.saveStopper--;
        while (cmp!=null) {
          if (cmp.saved>0) {
            cmp.saved--;
            insSingleStackOp(0, cmp.nr, 0, 0l, cmp.type);
          }
          //else restore without pop
          restoreRegs--;
          cmp=cmp.allocNext;
        }
        if (restoreRegs!=0) {
          ctx.out.println("SRB: invalid saveStopper (restore)");
          err=true;
        }
      }
    }
    //release killed register
    if (keepRegs!=deallocRegs && deallocRegs!=0) {
      cmp=lastReg;
      //search end of supporting register list
      while (cmp.type==T_SPRT && cmp.allocPrev!=null) cmp=cmp.allocPrev;
      //free register
      if (cmp.nr!=deallocRegs) {
        ctx.out.println("SRB: not last register in dealloc");
        err=true;
      }
      else if (cmp.saved>0 || cmp.saveStopper>0) {
        ctx.out.println("SRB: dealloc for saved/saveStopper register");
        err=true;
      }
      else {
        //everything ok, disable all helper registers
        while (cmp!=lastReg && lastReg!=null) disableLastRegister();
        //disable official register
        disableLastRegister();
      }
    }
  }
  
  public int getClssInstReg(boolean inst) {
    return inst ? R_INST : R_CLSS;
  }
  
  private void disableLastRegister() {
    currentContainer.closeReg(lastReg);
    lastReg=lastReg.allocPrev;
    if (lastReg!=null) lastReg.allocNext=null;
    //register is still in use in container, cleanup is done there
  }
  
  public void insMthdFrame(int enter, Mthd refMthd) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_MTHDFRAME;
    i.iPar1=enter;
    i.refMthd=refMthd;
    currentContainer.addBlock(i);
  }
  
  public void insFlowHint(int hint, int id) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_FLOWHINT;
    i.iPar1=hint;
    i.iPar2=id;
    currentContainer.addBlock(i);
  }
  
  public void insLoadConstVal(int dst, int ival, long lval, int type) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_LOADCONSTVAL;
    i.reg0=dst;
    i.iPar1=type;
    i.iPar2=ival;
    i.lPar=lval;
    currentContainer.addBlock(i);
  }
  
  public void insLoadVar(int loadValue, int dst, int src, int mode, int pos, int type) {
    Instruction i=getUnlinkedInstruction();
    switch (mode) {
      case M_STCK:
        src=R_BASE;
        break;
      case M_CLSS:
        src=R_CLSS;
        break;
      case M_INST:
        src=R_INST;
        break;
      case M_RESU:
        break;
      case M_ABSL:
        src=0;
        break;
      default:
        ctx.out.println("insLoadVar with invalid mode");
        err=true;
        return;
    }
    //mode is no longer required as src-register is correctly set now
    i.type=I_LOADVAR;
    i.reg0=dst;
    i.reg1=src;
    i.iPar1=type;
    i.iPar2=pos;
    i.iPar3=loadValue;
    currentContainer.addBlock(i);
  }
  
  public void insDupConv(int dst, int src, int toType, int fromType, int intfMapSpecial) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_DUPCONV;
    i.reg0=dst;
    i.reg1=src;
    i.iPar1=toType;
    i.iPar2=fromType;
    i.iPar3=intfMapSpecial;
    currentContainer.addBlock(i);
  }
  
  public void insSingleStackOp(int push, int reg, int ival, long lval, int type) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_SINGLESTACKOP;
    i.reg0=reg;
    i.iPar1=type;
    i.iPar2=push;
    i.iPar3=ival;
    i.lPar=lval;
    currentContainer.addBlock(i);
  }
  
  public void insAssign(int dst, int src, int type) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_ASSIGN;
    i.reg0=dst;
    i.reg1=src;
    i.iPar1=type;
    currentContainer.addBlock(i);
  }
  
  public void insRegOp(int dst, int src1, int src2, int op, int type) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_REGOP;
    i.reg0=dst;
    i.reg1=src1;
    i.reg2=src2;
    i.iPar1=type;
    i.iPar2=op;
    currentContainer.addBlock(i);
  }
  
  public void insMemOp(int dst, int op, int type) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_MEMOP;
    i.reg0=dst;
    i.iPar1=type;
    i.iPar2=op;
    currentContainer.addBlock(i);
  }
  
  public void insCall(int baseReg, int off, int parSize, Mthd mthd, int intf) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_CALL;
    i.reg0=baseReg;
    i.iPar1=parSize;
    i.iPar2=off;
    i.iPar3=intf;
    i.refMthd=mthd;
    currentContainer.addBlock(i);
  }
  
  public void insJump(Instruction dest, int cond) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_JUMP;
    i.jDest=dest;
    currentContainer.addBlock(i);
  }
  
  public void insCompare(int src1, int src2, int cond, int val, long lval, int type) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_COMPARE;
    i.reg0=src1;
    i.reg1=src2;
    i.iPar1=type;
    i.iPar2=cond;
    i.iPar3=val;
    i.lPar=lval;
    currentContainer.addBlock(i);
  }
  
  public void insSpecialMemIO(int write, int dst, int src, int type, int memLoc) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_SPECIALMEMIO;
    i.reg0=dst;
    i.reg1=src;
    i.iPar1=type;
    i.iPar2=write;
    i.iPar3=memLoc;
    currentContainer.addBlock(i);
  }
  
  public void insDerefAddr(int addrReg, int indReg, int baseOffset, int entrySize) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_DEREFADDR;
    i.reg0=addrReg;
    i.reg1=indReg;
    i.iPar1=baseOffset;
    i.iPar2=entrySize;
    currentContainer.addBlock(i);
  }

  public void insCheckBounds(int addrReg, int offReg, int checkToOffset, Instruction onSuccess) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_CHECKBOUNDS;
    i.reg0=addrReg;
    i.reg1=offReg;
    i.iPar1=checkToOffset;
    i.jDest=onSuccess;
    currentContainer.addBlock(i);
  }

  public void insNativeBoundException() {
    Instruction i=getUnlinkedInstruction();
    i.type=I_NATIVEBOUNDEXC;
    currentContainer.addBlock(i);
  }
  
  public void insThrowFrame(int iNr, int addrReg, Instruction dest, int throwBlockOffset) {
    Instruction i=getUnlinkedInstruction();
    i.type=I_THROWFRAME;
    i.reg0=addrReg;
    i.iPar1=iNr;
    i.iPar2=throwBlockOffset;
    i.jDest=dest;
    currentContainer.addBlock(i);
  }
  
  public void insInlineVarOffset(int inlineMode, int mode, int pos, int baseValue) {
    Instruction i=getUnlinkedInstruction();
    int src;
    switch (mode) {
      case M_STCK:
        src=R_BASE;
        break;
      case M_CLSS:
      case M_INST:
      case M_RESU:
        src=R_INST;
        break;
      case M_ABSL:
        src=0;
        break;
      default:
        ctx.out.println("insInlineVarOffset with invalid mode");
        err=true;
        return;
    }
    //src-register contains information of mode
    i.type=I_INLINEVAROFFSET;
    i.reg0=src;
    i.iPar1=inlineMode;
    i.iPar2=pos;
    i.iPar3=baseValue;
    currentContainer.addBlock(i);
  }

  public int insertFlowHint(int hint, int id) {
    SRBI_Container lastContainer;
    
    //handle container and insertion of hint
    switch (hint) {
      case F_BLOCKSTART:
        lastContainer=currentContainer;
        currentContainer=getUnlinkedContainer();
        lastContainer.addBlock(currentContainer);
        insFlowHint(hint, nextID++);
        break;
      case F_BLOCKEND:
        insFlowHint(hint, id);
        currentContainer=currentContainer.outer;
        break;
      default:
        insFlowHint(hint, id);
    }
    //depending on hint type, switch container state
    switch (hint) {
      case F_BLOCKSTART:
      case F_CONDSTART:
      case F_LOOPSTART:
      case F_SWSBSTART:
      case F_THRWMTHD:
      case F_TRUESTART:
      case F_ELSESTART:
      case F_TRYSTART:
      case F_CAFISTART:
        break;
    }
    //return id
    return id;
  }
  
  public void genLoadConstVal(int dst, int val, int type) {
    insLoadConstVal(dst, val, 0l, type);
  }
  
  public void genLoadConstDoubleOrLongVal(int dst, long val, boolean asDouble) {
    insLoadConstVal(dst, 0, val, asDouble ? StdTypes.T_DBL : StdTypes.T_LONG);
  }
  
  public void genLoadVarAddr(int dst, int src, int mode, int pos) {
    insLoadVar(0, dst, src, mode, pos, StdTypes.T_PTR);
  }
  
  public void genLoadVarVal(int dst, int src, int mode, int pos, int type) {
    insLoadVar(1, dst, src, mode, pos, type);
  }
  
  public void genConvertVal(int dst, int src, int toType, int fromType) {
    insDupConv(dst, src, toType, fromType, 0);
  }
  
  public void genDup(int dst, int src, int type) {
    insDupConv(dst, src, type, type, 0);
  }
  
  public void genPushConstVal(int val, int type) {
    insSingleStackOp(1, 0, val, 0l, type);
  }
  
  public void genPushConstDoubleOrLongVal(long val, boolean asDouble) {
    insSingleStackOp(1, 0, 0, val, asDouble ? StdTypes.T_DBL : StdTypes.T_LONG);
  }
  
  public void genPush(int src, int type) {
    insSingleStackOp(1, src, 0, 0l, type);
  }
  
  public void genPop(int dst, int type) {
    insSingleStackOp(0, dst, 0, 0l, type);
  }
  
  public void genAssign(int dst, int src, int type) {
    insAssign(dst, src, type);
  }
  
  public void genBinOp(int dst, int src1, int src2, int op, int type) {
    insRegOp(dst, src1, src2, op, type);
  }
  
  public void genUnaOp(int dst, int src, int op, int type) {
    insRegOp(dst, src, 0, op, type);
  }
  
  public void genIncMem(int dst, int type) {
    insMemOp(dst, 1, type);
  }
  
  public void genDecMem(int dst, int type) {
    insMemOp(dst, -1, type);
  }
  
  public void genSaveUnitContext() {
    insSingleStackOp(1, R_CLSS, 0, 0l, StdTypes.T_PTR);
  }
  
  public void genRestUnitContext() {
    insSingleStackOp(0, R_CLSS, 0, 0l, StdTypes.T_PTR);
  }
  
  public void genLoadUnitContext(int dst, int off, boolean replace) { //if replace, dst is not used
    if (replace) dst=R_CLSS;
    insLoadVar(1, dst, R_CLSS, M_RESU, off, StdTypes.T_PTR);
  }
  
  public void genLoadConstUnitContext(int dst, int unitAddr, boolean replace) { //as above
    if (replace) dst=R_CLSS;
    insLoadConstVal(dst, unitAddr, 0l, StdTypes.T_PTR);
  }
  
  public void genSaveInstContext() {
    insSingleStackOp(1, R_INST, 0, 0l, StdTypes.T_PTR);
    insSingleStackOp(1, R_CLSS, 0, 0l, StdTypes.T_PTR);
  }
  
  public void genRestInstContext() {
    insSingleStackOp(0, R_CLSS, 0, 0l, StdTypes.T_PTR);
    insSingleStackOp(0, R_INST, 0, 0l, StdTypes.T_PTR);
  }
  
  public void genLoadInstContext(int src) {
    insDupConv(R_INST, src, StdTypes.T_PTR, StdTypes.T_PTR, 0);
    insLoadVar(1, R_CLSS, R_INST, M_RESU, -relocBytes, StdTypes.T_PTR);
  }
  
  public void genCall(int off, int clssReg, boolean normal, int parSize) {
    if (normal) clssReg=R_CLSS;
    insCall(clssReg, off, parSize, null, 0);
  }
  
  public void genCallIndexed(int intfReg, int off, int parSize) {
    insCall(intfReg, off, parSize, null, 1);
  }
  
  public void genCallConst(Mthd obj, int parSize) {
    insCall(0, 0, parSize, obj, 0);
  }
  
  public void genJmp(Instruction dest) {
    insJump(dest, 0);
  }
  
  public void genCondJmp(Instruction dest, int condHnd) {
    insJump(dest, 0);
  }
  
  public int genComp(int src1, int src2, int type, int cond) {
    insCompare(src1, src2, cond, 0, 0l, type);
    return cond;
  }
  
  public int genCompValToConstVal(int src, int val, int type, int cond) {
    insCompare(src, 0, cond, val, 0l, type);
    return cond;
  }
  
  public int genCompValToConstDoubleOrLongVal(int src, long val, boolean asDouble, int cond) {
    insCompare(src, 0, cond, 0, val, asDouble ? StdTypes.T_DBL : StdTypes.T_LONG);
    return cond;
  }
  
  public int genCompPtrToNull(int reg, int cond) {
    insCompare(reg, 0, cond, 0, 0l, StdTypes.T_PTR);
    return cond;
  }
  
  public void genWriteIO(int dst, int src, int type, int memLoc) {
    insSpecialMemIO(1, dst, src, type, memLoc);
  }
  
  public void genReadIO(int dst, int src, int type, int memLoc) {
    insSpecialMemIO(0, dst, src, type, memLoc);
  }
  
  public void genMoveToPrimary(int src, int type) {
    insDupConv(R_PRIR, src, type, type, 0);
  }
  
  public void genMoveFromPrimary(int dst, int type) {
    insDupConv(dst, R_PRIR, type, type, 0);
  }
  
  public void genMoveIntfMapFromPrimary(int dst) {
    insDupConv(dst, R_PRIR, StdTypes.T_DPTR, StdTypes.T_DPTR, 1);
  }
  
  public void genSavePrimary(int type) {
    insSingleStackOp(1, R_PRIR, 0, 0l, type);
  }
  
  public void genRestPrimary(int type) {
    insSingleStackOp(0, R_PRIR, 0, 0l, type);
  }
  
  public void genLoadDerefAddr(int addrReg, int indReg, int baseOffset, int entrySize) {
    insDerefAddr(addrReg, indReg, baseOffset, entrySize);
  }
  
  public void genCheckBounds(int addrReg, int offReg, int checkToOffset, Instruction onSuccess) {
    insCheckBounds(addrReg, offReg, checkToOffset, onSuccess);
  }
  
  public void genNativeBoundException() {
    insNativeBoundException();
  }
  
  public void genThrowFrameBuild(int globalAddrReg, Instruction dest, int throwBlockOffset) {
    insThrowFrame(1, globalAddrReg, dest, throwBlockOffset);
  }
  
  public void genThrowFrameUpdate(Instruction oldDest, Instruction newDest, int throwBlockOffset) {
    insThrowFrame(2, 0, oldDest, throwBlockOffset);
    insThrowFrame(3, 0, newDest, throwBlockOffset);
  }
  
  public void genThrowFrameReset(int globalAddrReg, int throwBlockOffset) {
    insThrowFrame(4, globalAddrReg, null, throwBlockOffset);
  }

  public void inlineVarOffset(int inlineMode, int mode, int pos, int baseValue) {
    insInlineVarOffset(inlineMode, mode, pos, baseValue);
  }
  
  protected void printCode(TextPrinter v, String comment) {
    int insCnt;
    outestMthd.owner.printNameWithOuter(v);
    v.print('.');
    outestMthd.printNamePar(v);
    v.print(": //");
    v.println(comment);
    startContainer.enumerateInstructions(1);
    insCnt=startContainer.print(v);
    v.print("//instruction count ");
    v.print(comment);
    v.print(": ");
    v.println(insCnt);
  }
}
