/* Copyright (C) 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package backend.srb;

import osio.TextPrinter;
import backend.Instruction;

/**
 * SRBContainer: container for instructions and container 
 * 
 * @author S. Frenz
 * @version 090215 added getReg, initialization of parent and containerPrev setting
 *  version 090207 added copyright notice
 *  version 090118 added support for enumerateInstructions and print
 *  version 081226 initial version
 */

public class SRBI_Container extends SRBInstruction {
  private Instruction firstInstr, lastInstr;
  private SRBRegister regList;
  
  public SRBI_Container(SRB owner) {
    super(owner, 0); //do not allocate space for code-array in blocks
  }
  
  public void addBlock(Instruction i) {
    if (lastInstr==null) {
      lastInstr=firstInstr=i;
      i.prev=i.next=null;
      return;
    }
    lastInstr.next=i;
    i.prev=lastInstr;
    i.next=null;
    lastInstr=i;
    ((SRBInstruction)i).outer=this;
  }

  public void openReg(SRBRegister reg) {
    if ((reg.containerNext=regList)!=null) regList.containerPrev=reg;
    regList=reg;
    reg.isOpen=true;
  }
  
  public void closeReg(SRBRegister reg) {
    reg.isOpen=false;
  }
  
  public SRBRegister getReg(int regNr) {
    SRBRegister reg=regList;
    while (reg!=null) {
      if (reg.nr==regNr) return reg;
      reg=reg.containerNext;
    }
    if (outer!=null) return outer.getReg(regNr);
    return null;
  }
  
  public void resetMark() {
    Instruction tmp=firstInstr;
    while (tmp!=null) {
      ((SRBInstruction)tmp).resetMark();
      tmp=(SRBInstruction)tmp.next;
    }
  }
  
  public boolean isMarked() {
    Instruction tmp=firstInstr;
    boolean ret=false;
    while (tmp!=null) {
      ret|=((SRBInstruction)tmp).isMarked();
      tmp=(SRBInstruction)tmp.next;
    }
    return ret;
  }
  
  public void generateJumpDestsForTarget() {
    Instruction tmp=firstInstr;
    while (tmp!=null) {
      ((SRBInstruction)tmp).generateJumpDestsForTarget();
      tmp=tmp.next;
    }
  }
  
  public void convertToNative(int phase) {
    SRBRegister reg;
    Instruction tmp;
    //check if we have something to do
    switch (phase) {
      case SRB.P_RESOLVEREGISTERS:
        reg=regList;
        while (reg!=null) {
          owner.target.resolveRegister(reg);
          reg=reg.containerNext;
        }
        break;
      case SRB.P_MAPSUPPORTREGISTERS:
        reg=regList;
        while (reg!=null) {
          owner.target.mapNativeRegister(reg);
          reg=reg.containerNext;
        }
        break;
    }
    //call depending container
    tmp=firstInstr;
    while (tmp!=null) {
      ((SRBInstruction)tmp).convertToNative(phase);
      tmp=tmp.next;
    }
  }
  
  public void cleanUp() {
    Instruction tmpNext;
    while (firstInstr!=null) {
      tmpNext=firstInstr.next; //next is destroyed by cleanup, so remember real next
      ((SRBInstruction)firstInstr).cleanUp();
      firstInstr=tmpNext;
    }
    lastInstr=null;
    while (regList!=null) {
      regList.allocNext=owner.freeReg;
      owner.freeReg=regList;
      regList=regList.containerNext;
    }
    next=owner.freeContainer;
    owner.freeContainer=this;
  }
  
  public int enumerateInstructions(int startWith) {
    Instruction tmp=firstInstr;
    while (tmp!=null) {
      startWith=((SRBInstruction)tmp).enumerateInstructions(startWith);
      tmp=tmp.next;
    }
    return startWith;
  }
  
  public int print(TextPrinter v) {
    Instruction tmp=firstInstr;
    int cnt=0;
    while (tmp!=null) {
      cnt+=((SRBInstruction)tmp).print(v);
      tmp=tmp.next;
    }
    return cnt;
  }
}
