/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package frontend.sjava;

import compbase.*;
import debug.DebugWriter;

/**
 * JUnit: java-specific behaviour of units
 * 
 * @author S. Frenz
 * @version 090207 added copyright notice
 *  version 080703 adopted changed symInfo-debug-interface, added recursive parent-search
 *  version 080630 added search of inner unit in imported units 
 *  version 080624 added search of unit in outer class if existing
 *  version 080622 added support for static class initialization
 *  version 080616 added dependsOn-check for constant objects, added support for constant new
 *  version 080614 got strings and arrays from Clss as it may be used in Intf
 *  version 080506 added support for implicit conversion in class variable initialization
 *  version 080414 got instInitVars from Clss, adopted change in Mthd
 *  version 080122 moved $outer-checks to Clss
 *  version 080106 added support for direct non-interface descendants in variable initialization
 *  version 080105 added support for extended inner units
 *  version 071222 added support for inner units
 *  version 070909 optimized signature of Expr.resolve
 *  version 070903 adopted changes in AccVar.checkNameAgainst*
 *  version 070731 adopted change of QualID[] to QualIDList
 *  version 070628 adopted changes for "Expr extends TypeRef"
 *  version 070527 fixed unit variable location setting
 *  version 070501 moved prepareOutput to compbase.Unit
 *  version 070303 added support for movable indirect scalars
 *  version 070128 fixed static/dynamic check for initialized variables
 *  version 070127 optimized variable initialization, added initialization of instance variables 
 *  version 070114 reduced access level where possible
 *  version 070111 adapted change in printPos and compErr
 *  version 061203 optimized calls to printPos and compErr
 *  version 061027 added support for STRUCT
 *  version 060723 added support for extension of explicit standard constructor in parent
 *  version 060707 added type-check for class-variables
 *  version 060628 added support for static compilation
 *  version 060607 initial version
 */

public abstract class JUnit extends Unit {
  private final static int R_NOTSTARTED=0; //needs to be 0 for initialization of ...ResolveState
  private final static int R_STARTED   =1;
  private final static int R_SUCCESS   =2;
  private final static int R_ERROR     =3;
  
	//required fields for resolving
  protected QualIDList impt;
  protected VrblList instInitVars;
  protected ExStr strings;
  protected ExArrayInit arrays;
  protected ExConstNew constNews;
	private int intfResolveState, mthdResolveState; //...ResolveState==R_NOTSTARTED initialized automatically
	
	protected JUnit(int fid, int il, int ic) {
		super(fid, il, ic);
	}
	
	public boolean hasValidInterface() {
	  return intfResolveState==R_SUCCESS;
	}
	
	protected abstract boolean resolveIntfExtsIpls(Context ctx);
  protected abstract boolean checkInstVarInit(Vrbl var, Context ctx);
  protected abstract boolean resolveMthdExtsIpls(Context ctx);
	protected abstract boolean checkDeclarations(Context ctx);
	
  public Unit searchUnitInView(StringList what, boolean isRecursion) {
    Unit res;
    StringList cur;
    QualIDList list;
    
    //check ourself
    if (what.next==null && what.str.equals(name)) return this;
    //check inner units
    cur=what;
    if (cur.next!=null && cur.str.equals(name)) cur=cur.next; //full qualified type with ourself in front
    res=innerUnits;
    while (res!=null) {
      if (cur.str.equals(res.name)) { //found current inner unit
        if ((cur=cur.next)==null) return res; //found all
        res=res.innerUnits; //get next level
      }
      else res=res.nextUnit; //try next unit
    }
    //find in outer units
    if (outerUnit!=null) return outerUnit.searchUnitInView(what, false);
    if (!isRecursion) {
      //find in package
      if ((res=pack.packDest.searchUnit(what))!=null) return res;
      //find in imports
      list=impt;
      while (list!=null) {
        if (list.qid.type==QualID.Q_IMPORTPACK) { //import package
          if ((res=list.qid.packDest.searchUnit(what))!=null) return res;
        }
        else { //import unit
          if (what.str.equals(list.qid.unitDest.name)) {
            if (what.next==null) return list.qid.unitDest; //direct match
            //search inner units
            cur=what.next;
            res=list.qid.unitDest.innerUnits;
            while (res!=null) {
              if (cur.str.equals(res.name)) { //found current inner unit
                if ((cur=cur.next)==null) return res; //found all
                res=res.innerUnits; //get next level
              }
              else res=res.nextUnit; //try next unit
            }
          }
        }
        list=list.nextQualID;
      }
    }
    if (extsID!=null && extsID.unitDest!=null) return extsID.unitDest.searchUnitInView(what, true);
    //not found
    return null;
  }
  
	public boolean resolveInterface(Context ctx) {
	  boolean error=false;
	  Vrbl var;
	  Mthd mthd;
    QualIDList list;
	  
	  //check if already resolved
	  switch (intfResolveState) {
      case R_NOTSTARTED: break; //normal mode
      case R_STARTED: //this indicates cyclic call of CompileUnit.resolveInterface
        ctx.out.println("error: cyclic unit.resolveInterface - check extends- and implements-statements");
        ctx.out.print("in unit ");
  	    ctx.out.println(name);
  	    intfResolveState=R_ERROR;
  	    return false;
      case R_SUCCESS: return true; //already done, nothing to do
      case R_ERROR: return false; //silently abort, we printed an error-message already
      default: //this should never happen
        ctx.out.println("### invalid CompileUnit.intfResolveState ###");
        intfResolveState=R_ERROR;
        return false;
	  }
	  //start resolving
	  intfResolveState=R_STARTED;
    //test for special names
    if (SJava.isSpecialName(name) && this!=ctx.structClass) { //internally assigned STRUCT is valid
      printPos(ctx, "invalid use of keyword for class ");
      ctx.out.println(name);
      intfResolveState=R_ERROR;
      return false;
    }
	  //check if root entered
	  if (ctx.root==null) {
	    ctx.out.print("error: resolving invalid unit ");
	    ctx.out.print(name);
	    ctx.out.println(" not possible");
	    intfResolveState=R_ERROR;
	    return false;
	  }
	  //check import-statements
    list=impt;
    while (list!=null) {
      if (list.qid.type==QualID.Q_IMPORTPACK) list.qid.packDest=ctx.root.searchSubPackage(list.qid.name, false);
      else list.qid.unitDest=ctx.root.searchUnit(list.qid.name);
      if (list.qid.packDest==null && list.qid.unitDest==null) {
        list.qid.printPos(ctx, "could not resolve import-statement ");
        list.qid.printFullQID(ctx.out);
        ctx.out.print(" in unit ");
        ctx.out.println(name);
        error=true;
      }
      list=list.nextQualID;
    }
    //abort now on all of the following errors
	  if (error //abort if there was an error in import-statements
	      || !resolveIntfExtsIpls(ctx) //check extends and implements, has to be done differently in Clss and Intf
	      || (outerUnit!=null && !outerUnit.resolveInterface(ctx))) { //check outer if existing
	    intfResolveState=R_ERROR;
	    return false;
	  }
    //check methods
    mthd=mthds;
    while (mthd!=null) {
      if (!mthd.checkNameAndType(this, ctx)) {
        ctx.out.print(" in unit ");
        ctx.out.println(name);
        intfResolveState=R_ERROR;
        return false;
      }
      mthd=mthd.nextMthd;
    }
	  //check vars
	  var=vars;
	  while (var!=null) {
	    if (!var.type.resolve(this, ctx) || !var.checkNameAgainstVrbl(vars, ctx)
          || !var.checkNameAgainstUnits(innerUnits, ctx)
          || !var.enterSize(var.location==Vrbl.L_NOTRDY ? Vrbl.L_UNIT : var.location, ctx)
          || ((var.modifier&Vrbl.M_STAT)==0 && var.init!=null && !checkInstVarInit(var, ctx))) {
	      ctx.out.print(" in unit ");
	      ctx.out.println(name);
	      intfResolveState=R_ERROR;
	      return false;
	    }
	    var=var.nextVrbl;
	  }
	  //check MethodTable and check implementation of interfaces
	  if (!checkDeclarations(ctx)) {
	    intfResolveState=R_ERROR;
	    return false;
	  }
	  //everything OK
	  intfResolveState=R_SUCCESS;
	  return true;
	}
	
	public boolean resolveMethodBlocks(Context ctx) {
	  Vrbl var;
	  Mthd mthd;
	  boolean success=true;
	  int cmpRes;
	  ExEnc enc;
	  
    //to enable super-redirection and other optimizations, ensure to have parents resolved already
    switch (mthdResolveState) {
      case R_NOTSTARTED: break; //normal mode
      case R_STARTED: //this indicates cyclic call of CompileUnit.resolveMethodBlocks
        ctx.out.println("error: cyclic unit.resolveMethodBlocks - check extends- and implements-statements");
        ctx.out.print("in unit ");
        ctx.out.println(name);
        mthdResolveState=R_ERROR;
        return false;
      case R_SUCCESS: return true; //already done, nothing to do
      case R_ERROR: return false; //silently abort, we printed an error-message already
      default: //this should never happen
        ctx.out.println("### invalid CompileUnit.mthdResolveState ###");
        mthdResolveState=R_ERROR;
        return false;
    }
    mthdResolveState=R_STARTED;
    if (!resolveMthdExtsIpls(ctx)) {
      mthdResolveState=R_ERROR;
      return false;
    }
	  //initialization of variables belongs to code
	  var=vars;
	  while (var!=null) {
	    if ((var.modifier&Vrbl.M_FIN)==0) { //this is a not final variable (constants resolved seperately)
  	    if (var.init!=null) { //check if variable initialized
          mthd=(var.modifier&Vrbl.M_STAT)!=0 ? initStat : initDyna; //get static/dynamic context
          if (!var.init.resolve(this, mthd, ctx)) {
            ctx.out.print(" in unit ");
            ctx.out.println(name);
            success=false;
          }
          else if ((cmpRes=var.type.compareType(var.init, true, ctx))!=TypeRef.C_EQ
              && cmpRes!=TypeRef.C_TT) {
            if (cmpRes==TypeRef.C_NP || !var.type.isIntfType()) { //not implicitly convertible
              var.printPos(ctx, "incompatible or not implicitly convertible types in initialization of unit-variable");
              ctx.out.println();
              success=false;
            }
            else { //insert implicit conversion
              enc=new ExEnc(fileID, line, col);
              enc.convertTo=var.type;
              enc.ex=var.init;
              if (!enc.implConvResolve(this, ctx)) {
                ctx.out.print(" in inserted implicit conversion");
                return false;
              }
              var.init=enc;
            }
          }
        }
	    }
	    var=var.nextVrbl;
	  }
	  //resolve method-blocks
	  mthd=mthds;
	  while (mthd!=null) {
	    if (!mthd.resolve(this, instInitVars, ctx)) {
	      ctx.out.print(" in unit ");
	      ctx.out.println(name);
	      success=false;
	    }
	    mthd=mthd.nextMthd;
	  }
	  if ((initStat.modifier&Mthd.M_NDCODE)!=0) {
	    if (!initStat.resolve(this, null, ctx)) {
	      ctx.out.print(" in static init of unit ");
	      ctx.out.println(name);
	      success=false;
	    }
	  }
	  //done, success will be true if everything was resolved
    if (success) {
      mthdResolveState=R_SUCCESS;
      return true;
    }
    mthdResolveState=R_ERROR;
	  return false;
	}
	
	protected boolean resolveExtsIplsQID(QualID qid, Context ctx) { //called by Clss and TIntf only
	  //search for unit
	  if ((qid.unitDest=searchUnitInView(qid.name, false))!=null //use normal search function
	      || (qid.unitDest=ctx.defUnits.searchUnit(qid.name))!=null //search in default units
	      || (qid.unitDest=ctx.root.searchUnit(qid.name))!=null) {//search in root package
	    if (qid.unitDest==this) {
	      qid.printPos(ctx, "cannot extend/implement itself");
	      return false;
	    }
	    return true; //found destination
	  }
	  //not found
	  qid.printPos(ctx, "could not resolve ");
	  qid.printFullQID(ctx.out);
    ctx.out.print(" in unit ");
    ctx.out.println(name); 
	  return false;
	}
	
	protected void fixDynaAddresses() {
    ExStr checkStr;
    ExArrayInit checkArr;
    ExConstNew checkNew;
    
    if ((checkStr=strings)!=null | (checkArr=arrays)!=null | (checkNew=constNews)!=null) {
      modifier|=MA_ACCSSD;
      //enter strings
      while (checkStr!=null) {
        checkStr.target.varPos=++statRelocTableEntries+clssRelocTableEntries;
        checkStr.target.location=Vrbl.L_CLSSREL;
        checkStr=checkStr.nextStr;
      }
      //enter array initializations
      while (checkArr!=null) {
        checkArr.target.varPos=++statRelocTableEntries+clssRelocTableEntries;
        checkArr.target.location=Vrbl.L_CLSSREL;
        checkArr=checkArr.nextArrayInit;
      }
      //enter constant new objects
      while (checkNew!=null) {
        checkNew.target.varPos=++statRelocTableEntries+clssRelocTableEntries;
        checkNew.target.location=Vrbl.L_CLSSREL;
        checkNew=checkNew.nextConstNew;
      }
    }
	}

  public boolean genConstObj(Context ctx) {
    ExStr checkStr;
    ExArrayInit checkArr;
    ExConstNew checkNew;
    Unit unit;
    FilledParam pa;
    int cArAddr, off, baseType, extType, memoryHint;
    boolean objElements;
    
    //generate strings
    checkStr=strings;
    memoryHint=ctx.mem.getCurrentAllocAmountHint();
    while (checkStr!=null) {
      if (checkStr.dependsOn==null || (checkStr.dependsOn.modifier&Mthd.M_NDCODE)!=0) {
        ctx.stringCount++;
        ctx.stringChars+=checkStr.value.length();
        if ((checkStr.outputAddr=ctx.allocateString(checkStr.value))==0) {
          outputError=true;
          return false;
        }
        if (ctx.dynaMem) ctx.arch.putRef(outputAddr-checkStr.target.varPos*ctx.arch.relocBytes, checkStr.outputAddr);
        else checkStr.target.varPos=checkStr.outputAddr;
      }
      else if (ctx.verbose) {
        checkStr.printPos(ctx, "skipping string in unit ");
        ctx.out.println(name);
      }
      //get next string
      checkStr=checkStr.nextStr;
    }
    ctx.stringMemBytes+=ctx.mem.getCurrentAllocAmountHint()-memoryHint;
    //generate arrays
    checkArr=arrays;
    while (checkArr!=null) {
      if (checkArr.dependsOn==null || (checkArr.dependsOn.modifier&Mthd.M_NDCODE)!=0) {
        if (checkArr.qid!=null) {
          objElements=true;
          if (checkArr.qid.unitDest==null) {
            compErr(ctx, "invalid unitDest for array initialization");
            return false;
          }
          extType=checkArr.qid.unitDest.outputAddr;
        }
        else {
          objElements=checkArr.arrDim>1;
          extType=0;
        }
        baseType=checkArr.baseType;
        checkArr.outputAddr=cArAddr=ctx.mem.allocateArray(checkArr.len, checkArr.arrDim,
            objElements ? -1 : TypeRef.getMinSize(baseType), baseType, extType);
        if (cArAddr<0) {
          ctx.out.println("error allocating memory while creating constant array for initialization");
          outputError=true;
          return false;
        }
        //enter pointer
        if (ctx.dynaMem) ctx.arch.putRef(outputAddr-checkArr.target.varPos*ctx.arch.relocBytes, cArAddr);
        else if (ctx.embConstRAM) checkArr.target.varPos=cArAddr-ctx.ramInitAddr+ctx.ramStart;
        else checkArr.target.varPos=cArAddr;
        //enter values
        if (objElements) cArAddr-=(ctx.rteSArray.instRelocTableEntries+1)*ctx.arch.relocBytes;
        else {
          if (ctx.indirScalars) cArAddr=ctx.mem.getIndirAddrOfLastObject()+ctx.rteSArray.instIndirScalarTableSize;
          else cArAddr+=ctx.rteSArray.instScalarTableSize;
        }
        pa=checkArr.par;
        for (off=0; off<checkArr.len; off++) {
          if (pa==null) {
            ctx.out.println("internal error in parameter-count during array initialization");
            outputError=true;
            return false;
          }
          if (objElements) {
            if (pa.expr.baseType!=StdTypes.T_NULL) {
              if (!(pa.expr instanceof ExConstInitObj)) {
                compErr(ctx, "invalid initialization element in array initialization");
                outputError=true;
                return false;
              }
              ctx.arch.putRef(cArAddr-off*ctx.arch.relocBytes, ((ExConstInitObj)pa.expr).outputAddr);
            }
          }
          else switch (baseType) {
            case StdTypes.T_BOOL: case StdTypes.T_BYTE:
              ctx.mem.putByte(cArAddr+off, (byte)pa.expr.getConstIntValue(ctx));
              break;
            case StdTypes.T_SHRT: case StdTypes.T_CHAR:
              ctx.arch.putShort(cArAddr+off*2, (short)pa.expr.getConstIntValue(ctx));
              break;
            case StdTypes.T_INT: case StdTypes.T_FLT:
              ctx.arch.putInt(cArAddr+off*4, pa.expr.getConstIntValue(ctx));
              break;
            case StdTypes.T_LONG: case StdTypes.T_DBL:
              ctx.arch.putLong(cArAddr+off*8, pa.expr.getConstLongValue(ctx));
              break;
            default:
              compErr(ctx, "invalid base type for constant array");
              outputError=true;
              return false;
          }
          pa=pa.nextParam;
        }
      }
      else if (ctx.verbose) {
        checkArr.printPos(ctx, "skipping array in unit ");
        ctx.out.println(name);
      }
      //get next array
      checkArr=checkArr.nextArrayInit;
    }
    //generate objects
    checkNew=constNews;
    while (checkNew!=null) {
      unit=checkNew.destTypeUnit;
      if (checkNew.arrDim==0) {
        if ((checkNew.outputAddr=ctx.mem.allocate(unit.instScalarTableSize,
            unit.instIndirScalarTableSize, unit.instRelocTableEntries,
            unit.outputAddr))<0) {
          ctx.out.println("error allocating memory while creating pre-allocated object");
          outputError=true;
          return false;
        }
      }
      else {
        if (unit==null) {
          extType=0;
          objElements=false;
        }
        else {
          extType=unit.outputAddr;
          objElements=true;
        }
        if ((checkNew.outputAddr=ctx.mem.allocateMultiArray(checkNew.arrDim, checkNew.arrSizes,
            objElements ? -1 : TypeRef.getMinSize(checkNew.baseType),
                checkNew.baseType, extType, ctx))<0) {
          ctx.out.println("error allocating memory while creating pre-allocated array");
          outputError=true;
          return false;
        }
      }
      //enter pointer
      if (ctx.dynaMem) ctx.arch.putRef(outputAddr-checkNew.target.varPos*ctx.arch.relocBytes, checkNew.outputAddr);
      else if (ctx.embConstRAM) checkNew.target.varPos=checkNew.outputAddr-ctx.ramInitAddr+ctx.ramStart;
      else checkNew.target.varPos=checkNew.outputAddr;
      //get next new
      checkNew=checkNew.nextConstNew;
    }
    //everything done
    return true;
  }
  
  protected void writeCommonDebug(Context ctx, DebugWriter dbw) {
    Vrbl var;
    ExStr str;
    ExArrayInit arr;
    ExConstNew cnw;
    
    dbw.startVariableList();
    var=vars;
    while (var!=null) {
      dbw.hasVariable(var);
      var=var.nextVrbl;
    }
    dbw.endVariableList();
    dbw.startStatObjList();
    str=strings;
    while (str!=null) {
      if (str.dependsOn==null || (str.dependsOn.modifier&Mthd.M_NDCODE)!=0) {
        if (ctx.dynaMem) dbw.hasStatObj(-str.target.varPos*ctx.arch.relocBytes, str.outputAddr, str.value);
        else dbw.hasStatObj(-1, str.target.varPos, str.value);
      }
      else dbw.hasStatObj(-1, -1, str.value);
      str=str.nextStr;
    }
    arr=arrays;
    while (arr!=null) {
      if (arr.dependsOn==null || (arr.dependsOn.modifier&Mthd.M_NDCODE)!=0) {
        if (ctx.dynaMem) dbw.hasStatObj(-arr.target.varPos*ctx.arch.relocBytes, arr.outputAddr, null);
        else dbw.hasStatObj(-1, arr.target.varPos, null);
      }
      else dbw.hasStatObj(-1, -1, null);
      arr=arr.nextArrayInit;
    }
    cnw=constNews;
    while (cnw!=null) {
      if (ctx.dynaMem) dbw.hasStatObj(-cnw.target.varPos*ctx.arch.relocBytes, cnw.outputAddr, null);
      else dbw.hasStatObj(-1, cnw.target.varPos, null);
      //TODO cnw.printType(v);
      cnw=cnw.nextConstNew;
    }
    dbw.endStatObjList();
  }
}
